@@ -12,15 +12,15 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
-from __future__ import absolute_import
 import base64
+import functools
+import logging
 import threading
 import uuid
 
 import etcd3gw
 from etcd3gw import exceptions as etcd3_exc
 from oslo_utils import encodeutils
-import six
 
 import tooz
 from tooz import _retry
@@ -28,6 +28,8 @@ from tooz import coordination
 from tooz import locking
 from tooz import utils
 
+LOG = logging.getLogger(__name__)
+
 
 def _encode(data):
     """Safely encode data for consumption of the gateway."""
@@ -37,7 +39,7 @@ def _encode(data):
 def _translate_failures(func):
     """Translates common requests exceptions into tooz exceptions."""
 
-    @six.wraps(func)
+    @functools.wraps(func)
     def wrapper(*args, **kwargs):
         try:
             return func(*args, **kwargs)
@@ -161,7 +163,8 @@ class Etcd3Lock(locking.Lock):
         return False
 
 
-class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
+class Etcd3Driver(coordination.CoordinationDriverCachedRunWatchers,
+                  coordination.CoordinationDriverWithExecutor):
     """An etcd based driver.
 
     This driver uses etcd provide the coordination driver semantics and
@@ -169,15 +172,19 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
 
     The Etcd driver connection URI should look like::
 
-      etcd3+http://[HOST[:PORT]][?OPTION1=VALUE1[&OPTION2=VALUE2[&...]]]
+      etcd3+PROTOCOL://[HOST[:PORT]][?OPTION1=VALUE1[&OPTION2=VALUE2[&...]]]
 
-    If not specified, HOST defaults to localhost and PORT defaults to 2379.
+    The PROTOCOL can be http or https. If not specified, HOST defaults to
+    localhost and PORT defaults to 2379.
     Available options are:
 
     ==================  =======
     Name                Default
     ==================  =======
-    protocol            http
+    api_version         v3alpha
+    ca_cert             None
+    cert_key            None
+    cert_cert           None
     timeout             30
     lock_timeout        30
     membership_timeout  30
@@ -193,19 +200,35 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
     #: Default port used if none provided (4001 or 2379 are the common ones).
     DEFAULT_PORT = 2379
 
+    #: Default api version if none provided
+    DEFAULT_API_VERSION = "v3alpha"
+
     GROUP_PREFIX = b"tooz/groups/"
 
     def __init__(self, member_id, parsed_url, options):
         super(Etcd3Driver, self).__init__(member_id, parsed_url, options)
+        protocol = 'https' if parsed_url.scheme.endswith('https') else 'http'
         host = parsed_url.hostname or self.DEFAULT_HOST
         port = parsed_url.port or self.DEFAULT_PORT
         options = utils.collapse(options)
+        ca_cert = options.get('ca_cert')
+        cert_key = options.get('cert_key')
+        cert_cert = options.get('cert_cert')
         timeout = int(options.get('timeout', self.DEFAULT_TIMEOUT))
-        self.client = etcd3gw.client(host=host, port=port, timeout=timeout)
+        api_version = options.get("api_version", self.DEFAULT_API_VERSION)
+        self.client = etcd3gw.client(host=host,
+                                     port=port,
+                                     protocol=protocol,
+                                     ca_cert=ca_cert,
+                                     cert_key=cert_key,
+                                     cert_cert=cert_cert,
+                                     api_path="/" + api_version + "/",
+                                     timeout=timeout)
         self.lock_timeout = int(options.get('lock_timeout', timeout))
         self.membership_timeout = int(options.get(
             'membership_timeout', timeout))
         self._acquired_locks = set()
+        self._membership_lease = None
 
     def _start(self):
         super(Etcd3Driver, self)._start()
@@ -215,30 +238,22 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
         return Etcd3Lock(self, name, self.lock_timeout)
 
     def heartbeat(self):
+        # TODO(kaifeng) use the same lease for locks?
+        if self._membership_lease.refresh() == -1:
+            expired_lease = self._membership_lease.id
+            self._membership_lease = self.client.lease(self.membership_timeout)
+            LOG.debug('Created new lease %s after previous lease %s expired.',
+                      self._membership_lease.id, expired_lease)
         # NOTE(jaypipes): Copying because set can mutate during iteration
         for lock in self._acquired_locks.copy():
             lock.heartbeat()
-        # TODO(kaifeng) use the same lease for locks?
-        self._membership_lease.refresh()
         return min(self.lock_timeout, self.membership_timeout)
 
-    def watch_join_group(self, group_id, callback):
-        raise tooz.NotImplemented
-
-    def unwatch_join_group(self, group_id, callback):
-        raise tooz.NotImplemented
-
-    def watch_leave_group(self, group_id, callback):
-        raise tooz.NotImplemented
-
-    def unwatch_leave_group(self, group_id, callback):
-        raise tooz.NotImplemented
-
     def _encode_group_id(self, group_id):
         return _encode(self._prefix_group(group_id))
 
     def _prefix_group(self, group_id):
-        return b"%s%s/" % (self.GROUP_PREFIX, group_id)
+        return b"%s%s/" % (self.GROUP_PREFIX, utils.to_binary(group_id))
 
     def create_group(self, group_id):
         @_translate_failures
@@ -423,3 +438,11 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
                 group[1]['key'][len(self.GROUP_PREFIX):-1] for group in groups]
         return coordination.CoordinatorResult(
             self._executor.submit(_get_groups))
+
+    @staticmethod
+    def watch_elected_as_leader(group_id, callback):
+        raise tooz.NotImplemented
+
+    @staticmethod
+    def unwatch_elected_as_leader(group_id, callback):
+        raise tooz.NotImplemented
